/* $Id: main.cc,v 1.17 2001/05/15 08:32:58 ltd Exp $ */
#include <std.h>
#include <strings.h>
#include <ops.h>
#include <stdio.h>

#include <func/apt.h>
#include <func/config.h>
#include <func/keys.h>
#include <func/mouse.h>
#include <func/dependency.h>
#include <func/statecache.h>
#include <list/screens.h>
#include <interface/deityui.h>
#include <backend/acquire.h>

#include <getopt.h>

#ifndef DEFAULTUI
#define DEFAULTUI "curses"
#endif
#ifndef UIPREFIX
#define UIPREFIX "deity-"
#endif

DeityUI *CaptUI = NULL;

static const char *progname;

void initarguments(int, char **);
void InvalidOpt(char *, int);
void DisplayVersion();
void DisplayHelp(char *);
void sighandler(int);
void io_block_loop(void);

/*
 * Wrapper function to start up a ui based on various settings
 */
DeityUI *CreateUI(int &argc, char **&argv)
{
	DeityUI *ui = NULL;
	const char *name = NULL;

	name = getenv("CAPTUI");

	if (name == NULL)
	{
		name = strrchr(progname, '/');
		if (name == 0)
			name = progname;
		else
			name++;

		if (strncmp(name, UIPREFIX, sizeof UIPREFIX - 1) == 0)
			name += (sizeof UIPREFIX - 1);
		else
			name = DEFAULTUI;
	}

	string s;

	s = MODULE_PATH;
	if (s[s.length()-1] != '/')
		s += '/';
	s += name;

	INFO(INFO_DEBUG, "Selected interface: %s\n", s.c_str());
	ui = DeityUI::CreateInterface(s.c_str(), argc, argv);

	if (ui == NULL)
	{
		s = MODULE_PATH;
		if (s[s.length()-1] != '/')
			s += '/';
		s += DEFAULTUI;
		ui = DeityUI::CreateInterface(s.c_str(), argc, argv);
	}

	return ui;
}

int main(int argc, char **argv)
{
	// Set up the signal handlers
	signal(SIGSEGV, sighandler);
	signal(SIGINT, sighandler);
	signal(SIGKILL, sighandler);
	signal(SIGTERM, sighandler);

	progname = argv[0];

	// Initialize the apt config structures
	if (!pkgInitConfig(*_config) || !pkgInitSystem(*_config, _system))
	{
		_error->DumpErrors();
		return 100;
	}

	// Initialize command line arguments
	initarguments(argc, argv);

	// Parse the configuration file
	if (_Conf.Init() == false)
		_Conf.Write();

	// Initialize the user interface
	if ((CaptUI = CreateUI(argc, argv)) == NULL)
	{
		fprintf(stderr, "\nCould not load any UI modules. Try setting your dynamic loading path.\n\nFor example:  LD_LIBRARY_PATH=/usr/lib/deity/modules capt\nor load capt using the deity-<ui> symlinks.\n\n");
		exit(-4);
	}

	// Initialise the APT component
	Cache = new deitypkgcache;
	check_dependencies(false);

	// Check that there are no apt-generated errors at this point
	while (_error->PendingError() == true)
	{
		string s;
		_error->PopMessage(s);
		CaptUI->ShowDialog("Startup failure", (char *) s.c_str());
	}

	// Add the first screen
	CaptUI->scrlst.add();
	CURRENTSCR->BoundsReset();

	// Turn on description window
	if (_Conf.GetBool("deity.description.show") == true)
		CURRENTSCR->DescWindowShow();

	int UIret = CaptUI->Run();

	deity_clean_exit();

	return UIret;
}

void initarguments(int argc, char **argv)
{
	if (argc == 1)
		return;

	int fd;
	int c;
	int PreparedPkgs = -1;
	TextOp::opType CurrentT = TextOp::none;
	TextOp::CacheInit = false;
	TextOp::opRealCount = 0;

	for (c = 1; c < argc; c++)
	{
		if (*argv[c] == '-')
		{
			if (PreparedPkgs == 0)
			{
				fprintf(stderr, "option '%s' requires at least one argument [line opt %d]\n", argv[c - 1], c - 1);
				delete CaptUI;
				exit(255);
			}

			// These ones exit the program

			if (strcmp(argv[c] + 1, "version") == 0 || strcmp(argv[c] + 1, "v") == 0)
				DisplayVersion();
			else if (strcmp(argv[c] + 1, "help") == 0 || strcmp(argv[c] + 1, "h") == 0)
				DisplayHelp(argv[0]);

			// These don't :)

			else if (strcmp(argv[c] + 1, "install") == 0 || strcmp(argv[c] + 1, "i") == 0)
			{
				CurrentT = TextOp::install;
				PreparedPkgs = 0;
			}
			else if (strcmp(argv[c] + 1, "delete") == 0 || strcmp(argv[c] + 1, "d") == 0)
			{
				CurrentT = TextOp::remove;
				PreparedPkgs = 0;
			}
			else if (strcmp(argv[c] + 1, "purge") == 0 || strcmp(argv[c] + 1, "p") == 0)
			{
				CurrentT = TextOp::purge;
				PreparedPkgs = 0;
			}
			else if (strcmp(argv[c] + 1, "update") == 0 || strcmp(argv[c] + 1, "u") == 0)
			{
				CurrentT = TextOp::update;

				if (TextOp::CacheInit == false)
				{
					CaptUI = CreateUI(argc, argv);
					Cache = new deitypkgcache;
				}

				pkgAcquireBackend *Fetcher = new pkgAcquireBackend(CaptUI);
				fd = Fetcher->AcquireUpdate();
				CaptUI->WaitForFetch(Fetcher, fd);
				CaptUI->Run();
				delete Fetcher;

				TextOp::CacheInit = false;
				PreparedPkgs = -1;
			}
			else if (strcmp(argv[c] + 1, "upgrade") == 0 || strcmp(argv[c] + 1, "g") == 0)
			{
				CurrentT = TextOp::upgrade;

				if (TextOp::CacheInit == false)
				{
					CaptUI = CreateUI(argc, argv);
					Cache = new deitypkgcache;
				}

				pkgAcquireBackend *Fetcher = new pkgAcquireBackend(CaptUI);
				fd = Fetcher->AcquireUpgrade();
				CaptUI->WaitForFetch(Fetcher, fd);
				CaptUI->Run();
				delete Fetcher;

				TextOp::CacheInit = false;
				PreparedPkgs = -1;
			}
			else
				InvalidOpt(argv[c] + 1, c);
		}
		else
		{
			switch (CurrentT)
			{
			case TextOp::none:
			case TextOp::update:
			case TextOp::upgrade:
				InvalidOpt(argv[c], c);
				break;
			default:
				TextOp::PrepareOp(CurrentT, (string) argv[c]);
				PreparedPkgs++;
				break;
			}
		}
	}

	if (CurrentT != TextOp::none)
	{
		if (PreparedPkgs > 0 && TextOp::opRealCount > 0)
		{
			if (TextOp::CacheInit == false)
			{
				CaptUI = CreateUI(argc, argv);
				Cache = new deitypkgcache;
			}

			pkgAcquireBackend *Fetcher = new pkgAcquireBackend(CaptUI);
			fd = Fetcher->AcquireRun();
			CaptUI->WaitForFetch(Fetcher, fd);
			CaptUI->Run();
			delete Fetcher;
			delete CaptUI;
			exit(0);
		}
		else if (PreparedPkgs == 0)
		{
			fprintf(stderr, "'%s': not enough args [line opt %d]\n", argv[c - 1], c - 1);
			exit(255);
		}
		else
		{
			delete CaptUI;
			exit(10);
		}
	}
}

void InvalidOpt(char *optname, int optnum)
{
	fprintf(stderr, "syntax error at '%s' [line opt %d]\n", optname, optnum);
	exit(255);
}


void DisplayVersion()
{
	printf("Console (Advanced) Acquisition Package Tool (capt).\n"
		"Created and maintained by:\n"
		"  Patrick Cole   (ltd@debian.org)\n"
		"  Randolph Chung (tausq@debian.org)\n"
		"Compiled on %s %s.\nInternal Project Version String: %s %s\n", __TIME__, __DATE__, PACKAGE, VERSION);

	exit(0);
}

void DisplayHelp(char *caller)
{
#define _A_BOLD   "[1m"
#define _A_NORM   "[0m"

	printf("\nUsage: %s -(op 1) [package1 .. package2 .. packageN] ... -(op N) [pkg1 .. pkgN]\n\n", caller);
	printf(_A_BOLD "Information about this program\n" _A_NORM
				 "\n"
	       " -<" _A_BOLD "h" _A_NORM ">elp                    This helpful little screen\n"
	       " -<" _A_BOLD "v" _A_NORM ">ersion                 Current version of the program\n"
				 "\n"
				 _A_BOLD "Modify the state of a given set of packages\n" _A_NORM
				 "\n"
	       " -<" _A_BOLD "d" _A_NORM ">elete <pkg1..N>        Remove package(s)\n"
	       " -<" _A_BOLD "p" _A_NORM ">urge <pkg1..N>         Remove package(s) and config files\n"
				 "\n"
				 _A_BOLD "Acquire items from your defined archive sites and/or modify state\n" _A_NORM
				 "\n"
	       " -<" _A_BOLD "i" _A_NORM ">nstall <pkg1..N>       Install package(s)\n"
	       " -<" _A_BOLD "s" _A_NORM ">ource <pkg1..N>        Acquire source tarball for package(s)\n"
	       " -<" _A_BOLD "u" _A_NORM ">pdate-lists            Acquire new package listings if available\n"
				 "\n"
				 _A_BOLD "Auto dependency resolving upgrade features\n" _A_NORM
				 "\n"
	       " -up<" _A_BOLD "g" _A_NORM ">rade-only            Perform upgrade of installed packages\n\n"
	       " -<" _A_BOLD "d" _A_NORM ">ist-upgrade            Perform intelligent upgrade of packages\n"
	       "                                                  (may result in the removal of packages)\n\n");

	exit(0);
}

void sighandler(int sig)
{
	deity_clean_exit();
	fprintf(stderr, "Caught Signal %d\n", sig);
	if (sig == 11)
		fprintf(stderr, "Report this bug to deity@packages.debian.org\n");
	exit(sig);
}

void deity_clean_exit(void)
{
	if (Cache)
		delete Cache;

	if (CaptUI)
		delete CaptUI;

	_Conf.Write();
}
